#include <msp430x14x.h>

#include "clock.h"
#include "common.h"
#include "serial.h"
#include "sensors.h"

#define SDA_HIGH  (P5OUT |= 0x01)
#define SDA_LOW   (P5OUT &= 0xFE)
#define SCL_HIGH  (P5OUT |= 0x02)
#define SCL_LOW   (P5OUT &= 0xFD)

void initClock(void){ 
    setCSReg(0xC2);                       // signal output high on alarm (2) 
    P1IES &= 0xDF;                        // 0 - irflag on low to high   
    P1IE  |= 0x20;                        // 1 - enable clock interrupt 
}

//++++++++++++++++++++++++++++++++ Internal Bus Operations +++++++++++++++++++++++++++++++++++++++

// pre: bus idle => SDA SCL high
// post: 
// put start condition on bus, pre: clockSDA clockSCL output
void cl_start(void){
  SCL_LOW;  _NOP(); _NOP();
  SDA_HIGH; _NOP(); _NOP();
  SCL_HIGH; _NOP(); _NOP();
  SDA_LOW;  _NOP(); _NOP();
  SCL_LOW;  _NOP(); _NOP();
}

// put stop condition on bus, pre: clockSDA clockSCL output
void cl_stop(void){
  SCL_LOW;  _NOP(); _NOP();
  SDA_LOW;  _NOP(); _NOP();
  SCL_HIGH; _NOP(); _NOP();
  SDA_HIGH; _NOP(); _NOP();
  SCL_LOW;  _NOP(); _NOP();
}

// returns zero if failed
// pre:  SCL_LOW
int cl_writeOnBus(char byte) {
  int i, ack;
  for(i=0;i<8;i++) {
    if(byte & 0x80) SDA_HIGH; else SDA_LOW;
    SCL_HIGH;
    byte = byte << 1;
    SCL_LOW;
  }
  // check ack
  P5DIR &= 0xFE;                              // P50(SDA) input
  SCL_HIGH;
  if(P5IN & 0x01) ack = 0; else ack = 1;      // test if ack=0, else error
  SCL_LOW;
  P5DIR |= 0x01;                              // P50(SDA) output
  return ack;
}

// pre:  SCL_LOW
char cl_readFromBus(int ack) {
  int i;
  char byte = 0;
  P5DIR &= 0xFE;                              // P50(SDA) input
  for(i=0;i<8;i++) {
    byte = byte << 1;                
    SCL_HIGH;
    if(P5IN & 0x01) byte |= 0x01; else byte &= 0xFE;
    SCL_LOW;
  }
  P5DIR |= 0x01;                              // P50(SDA) output 
  if(ack) SDA_LOW; else SDA_HIGH;
  SCL_HIGH;
  SCL_LOW;
  return byte;
}

////////////////////////////////////////////////////////////////////////////////
//+++++++++++++++++++++++++++++++++++++++++++++ methods for TimeRecords (bjrn)
////////////////////////////////////////////////////////////////////////////////

void createTimeRecord(struct timeRecord* t){
  unsigned char reg;

  do cl_start();                    // do start until BUS_WRITE is acked
  while(!cl_writeOnBus(BUS_WRITE)); // control byte
  cl_writeOnBus(ACC_CLOCK);         // command byte
  cl_writeOnBus(0x00);              // starting address

  cl_start(); 
  cl_writeOnBus(BUS_READ);

  reg = cl_readFromBus(1);
  (*t).tseconds = reg >> 4;
  (*t).seconds  = reg;

  reg = cl_readFromBus(1);
  (*t).tminutes = reg >> 4;
  (*t).minutes  = reg;

  reg = cl_readFromBus(1);
  (*t).thours = reg >> 4;
  (*t).hours  = reg;

  reg = cl_readFromBus(1); // day of week unimportant

  reg = cl_readFromBus(1);
  (*t).tday = reg >> 4;
  (*t).day  = reg;

  reg = cl_readFromBus(1);
  (*t).tmonth = reg >> 4;
  (*t).month  = reg;

  reg = cl_readFromBus(0);
  (*t).tyear = reg >> 4;
  (*t).year  = reg;
  
  cl_stop();
}

void setTimeRecord(struct timeRecord* t){
  do cl_start();                    // do start until BUS_WRITE is acked
  while(!cl_writeOnBus(BUS_WRITE)); // control byte
  cl_writeOnBus(ACC_CLOCK);         // command byte
  cl_writeOnBus(0x00);              // starting address

  cl_writeOnBus( (((*t).tseconds)<<4) | ((*t).seconds)  );
  cl_writeOnBus( (((*t).tminutes)<<4) | ((*t).minutes)  );
  cl_writeOnBus( (((*t).thours)<<4) | ((*t).hours)  );
  cl_writeOnBus( 0x00 );
  cl_writeOnBus( (((*t).tday)<<4) | ((*t).day)  );
  cl_writeOnBus( (((*t).tmonth)<<4) | ((*t).month)  );
  cl_writeOnBus( (((*t).tyear)<<4) | ((*t).year)  );
  cl_stop();
}

// post: c = "dd.mm.yy hh:mm:ss"
//            01234567890123456
void timeRecordToString(struct timeRecord* t, char* c){
  c[0] = 0x30 + (*t).tday;
  c[1] = 0x30 + (*t).day;
  c[2] = '.';
  c[3] = 0x30 + (*t).tmonth;
  c[4] = 0x30 + (*t).month;
  c[5] = '.';
  c[6] = 0x30 + (*t).tyear;
  c[7] = 0x30 + (*t).year;
  c[8] = ' ';
  c[9] = 0x30 + (*t).thours;
  c[10] = 0x30 + (*t).hours;
  c[11] = ':';
  c[12] = 0x30 + (*t).tminutes;
  c[13] = 0x30 + (*t).minutes;
  c[14] = ':';
  c[15] = 0x30 + (*t).tseconds;
  c[16] = 0x30 + (*t).seconds;
  c[17] = 0x00;
}

void timeRecordFromString(struct timeRecord* t, char* c){
  (*t).tday      = c[0] - 0x30;
  (*t).day       = c[1] - 0x30;
  (*t).tmonth    = c[3] - 0x30;
  (*t).month     = c[4] - 0x30;
  (*t).tyear     = c[6] - 0x30;
  (*t).year      = c[7] - 0x30;
  (*t).thours    = c[9] - 0x30;
  (*t).hours     = c[10] - 0x30;
  (*t).tminutes  = c[12] - 0x30;
  (*t).minutes   = c[13] - 0x30;
  (*t).tseconds  = c[15] - 0x30;
  (*t).seconds   = c[16] - 0x30;
}

// compares to timestamps, returns 1 if t2 is greater ( newer ) than t1 
unsigned char compareTimeStamps(struct timeRecord* t1, struct timeRecord* t2){
  if( t2->tyear    != t1->tyear ) return (t2->tyear > t1->tyear );
  if( (*t2).year     != (*t1).year )     return ((*t2).year     > (*t1).year );
  if( (*t2).tmonth   != (*t1).tmonth )   return ((*t2).tmonth   > (*t1).tmonth );
  if( (*t2).month    != (*t1).month )    return ((*t2).month    > (*t1).month );
  if( (*t2).tday     != (*t1).tday )     return ((*t2).tday     > (*t1).tday );
  if( (*t2).day      != (*t1).day )      return ((*t2).day      > (*t1).day );
  if( (*t2).thours   != (*t1).thours )   return ((*t2).thours   > (*t1).thours );
  if( (*t2).hours    != (*t1).hours )    return ((*t2).hours    > (*t1).hours );
  if( (*t2).tminutes != (*t1).tminutes ) return ((*t2).tminutes > (*t1).tminutes );
  if( (*t2).minutes  != (*t1).minutes )  return ((*t2).minutes  > (*t1).minutes );
  if( (*t2).tseconds != (*t1).tseconds ) return ((*t2).tseconds > (*t1).tseconds );
  return ((*t2).seconds > (*t1).seconds );  
}


long int timeStampDiff(struct timeRecord* t1, struct timeRecord* t2){
  long int diffs = 0, t1s = 0, t2s = 0;    

  t1s = (t1->seconds) + (t1->tseconds * 10) + (t1->minutes*60) + (t1->tminutes*600) +
        (t1->hours*3600) + (t1->thours*36000) + (t1->day*86400) + (t1->tday*864000) +
        ((t1->month + t1->tmonth*10)*86400*31);

  t2s = (t2->seconds) + (t2->tseconds * 10) + (t2->minutes*60) + (t2->tminutes*600) +
        (t2->hours*3600) + (t2->thours*36000) + (t2->day*86400) + (t2->tday*864000) +
        ((t2->month + t2->tmonth*10)*86400*31);
  diffs = t2s - t1s;
  return diffs;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////
//+++++++++++++++++++ Several used Methods Read/Write from/to a special register +++++++++++++++++++++++++


// returns specified byte in the specified Control Register
// parameters:  acc - control register, adr - byte address , mask - last n bits of specified byte
char getRegEntry(char acc, char adr, char mask) {
  char value = 0;
  cl_start();
  cl_writeOnBus(BUS_WRITE);
  cl_writeOnBus(acc);
  cl_writeOnBus(adr);
  cl_start();
  cl_writeOnBus(BUS_READ);
  value = cl_readFromBus(0);
  value &= mask;               // choose last n bit
  cl_stop();
  return value;
}

// sets specified byte in the specified Control Register
// parameters:  acc - control register, adr - byte address , mask - last n bits of specified byte, data
void setRegEntry(char acc, char adr, char mask, char data) {
  do cl_start(); 
  while(!cl_writeOnBus(BUS_WRITE));
  cl_writeOnBus(acc);
  cl_writeOnBus(adr);
  data &= mask;               // choose last n bit
  cl_writeOnBus(data);
  cl_stop();
}

// reads out all bits of a 16 bit register (possible access: Configuration Status Register, Temperature, TH, TL)
int getReg16bit(char acc,int bitmask){
  int config = 0;
  do cl_start();
  while(!cl_writeOnBus(BUS_WRITE)); 
  cl_writeOnBus(acc);
  cl_start(); 
  cl_writeOnBus(BUS_READ);
  config = cl_readFromBus(1);
  config = config << 8;
  config += cl_readFromBus(0);
  cl_stop();
  config &= bitmask;
  _NOP();
  _NOP();
  return config;
}

// only first 8 bit of Configuration Status Register can be set
void setCSReg(char setting){
  do cl_start();
  while(!cl_writeOnBus(BUS_WRITE));               
  cl_writeOnBus(ACC_CSR);                         
  cl_writeOnBus(setting);
  cl_stop();
  _NOP();
  _NOP();
  _NOP();
  _NOP();
}

//++++++++++++++++++++++++++++++++++++ Clock SRAM Access ++++++++++++++++++++++++++++++++++++++++

void storeByteCSRAM(char adr, char data) {
  setRegEntry(ACC_CSRAM, adr, data, 0xFF);
}

char loadByteCSRAM(char adr) {
  return getRegEntry(ACC_CSRAM, adr, 0xFF);
}

//++++++++++++++++++++++++++++++++ Temperature Operations +++++++++++++++++++++++++++++++++++++++


int convertTemperature(char* c) {
  int value;
  value = (c[1] - 0x30) * 100 + (c[2] - 0x30) * 10 + (c[3] - 0x30);
  if(c[0] == '-') value = value * -1;
  value = (value << 8);
  if(c[5] == '5') value += 0x80;
  return value;
}

void loadTemperature(char* temp, int value) {
  char nk = (value & 0x0080);
  char vk = (value >> 8);
  if(vk < 0) temp[0] = '-'; else temp[0] = '+';
  if(vk & 0x80) vk = ~vk;  
  temp[4]='.';
  temp[3] = 0x30 + (vk % 10);
  vk = vk / 10;
  temp[2] = 0x30 + (vk % 10);
  vk = vk / 10;
  temp[1] = 0x30 + vk;  
  if(nk) temp[5] = '5'; else temp[5] = '0';
} 

void setThreshold(char acc, int temp) {
  do cl_start();
  while(! cl_writeOnBus(BUS_WRITE));          // opcode:  
  cl_writeOnBus(acc);                         // command:  
  cl_writeOnBus(temp >> 8);
  cl_writeOnBus(temp);
  cl_stop();
  _NOP();
  _NOP();
}

// set threshold (high)
void setTemperatureThresholdHigh(char* tt_high){
  setThreshold(ACC_TH, convertTemperature(tt_high));
}

// set threshold (low)
void setTemperatureThresholdLow(char* tt_low) {
  setThreshold(ACC_TL, convertTemperature(tt_low));
}

void readTemperatureThresholdHigh(char* adr){
  loadTemperature(adr,getReg16bit(ACC_TH,0xFFFF));
}

void readTemperatureThresholdLow(char* adr){
  loadTemperature(adr,getReg16bit(ACC_TL,0xFFFF));
}

void readTemperature(char* adr) {
  loadTemperature(adr,getReg16bit(ACC_RT,0xFFFF));
}

//++++++++++++++++++++++++++++++++  Clock and ClockAlarm Operations +++++++++++++++++++++++++++++++++++++++++

char hexToChar(int p, char c) {
  if(p) return (c >> 4) + 0x30; else return (0x30 + (c & 0x0F));
}

char getSeconds(void)         { return getRegEntry(ACC_CLOCK, ADR_SECONDS, MASK_SECONDS); }
char getMinutes(void)         { return getRegEntry(ACC_CLOCK, ADR_MINUTES, MASK_MINUTES); }
char getHours(void)           { return getRegEntry(ACC_CLOCK, ADR_HOURS, MASK_HOURS); }
char getWeekday(void)         { return getRegEntry(ACC_CLOCK, ADR_WEEKDAY, MASK_WEEKDAY); }
char getDay(void)             { return getRegEntry(ACC_CLOCK, ADR_DAY, MASK_DAY); }
char getMonth(void)           { return getRegEntry(ACC_CLOCK, ADR_MONTH, MASK_MONTH); }
char getYear(void)            { return getRegEntry(ACC_CLOCK, ADR_YEAR, MASK_YEAR); }

char getAlarmSeconds(void)    { return getRegEntry(ACC_CLOCK_ALARM, ADR_SECONDS, MASK_SECONDS);}
char getAlarmMinutes(void)    { return getRegEntry(ACC_CLOCK_ALARM, ADR_MINUTES, MASK_MINUTES);}
char getAlarmHours(void)      { return getRegEntry(ACC_CLOCK_ALARM, ADR_HOURS, MASK_HOURS);}
char getAlarmWeekday(void)    { return getRegEntry(ACC_CLOCK_ALARM, ADR_WEEKDAY, MASK_WEEKDAY);}  


void setDate(char* date) {
  setRegEntry(ACC_CLOCK, ADR_DAY,   MASK_DAY, ((date[0] & 0x0F) << 4) + (date[1] & 0x0F));
  setRegEntry(ACC_CLOCK, ADR_MONTH, MASK_MONTH, ((date[3] & 0x0F) << 4) + (date[4] & 0x0F));
  setRegEntry(ACC_CLOCK, ADR_YEAR, MASK_YEAR,   ((date[6] & 0x0F) << 4) + (date[7] & 0x0F));
  setRegEntry(ACC_CLOCK, ADR_WEEKDAY, MASK_WEEKDAY, date[9] & 0x0F);   
}

void setTime(char* time) {
  setRegEntry(ACC_CLOCK, ADR_HOURS,   MASK_HOURS,   ((time[0] & 0x0F) << 4) + (time[1] & 0x0F));
  setRegEntry(ACC_CLOCK, ADR_MINUTES, MASK_MINUTES, ((time[3] & 0x0F) << 4) + (time[4] & 0x0F));
  setRegEntry(ACC_CLOCK, ADR_SECONDS, MASK_SECONDS, ((time[6] & 0x0F) << 4) + (time[7] & 0x0F));
}

// set alarm (convert extended char-time to intern used hex-time-mode)
void setAlarmTime(char* ext_time) {
  setRegEntry(ACC_CLOCK_ALARM, ADR_HOURS,   MASK_HOURS,   ((ext_time[0] & 0x0F) << 4) + (ext_time[1] & 0x0F));
  setRegEntry(ACC_CLOCK_ALARM, ADR_MINUTES, MASK_MINUTES, ((ext_time[3] & 0x0F) << 4) + (ext_time[4] & 0x0F));
  setRegEntry(ACC_CLOCK_ALARM, ADR_SECONDS, MASK_SECONDS, ((ext_time[6] & 0x0F) << 4) + (ext_time[7] & 0x0F));
  setRegEntry(ACC_CLOCK_ALARM, ADR_WEEKDAY, MASK_WEEKDAY, ext_time[9] & 0x0F);
}

void readDate(char* adr) {
  char c = getDay();
  adr[0] = hexToChar(1,c);
  adr[1] = hexToChar(0,c);
  adr[2] = '-';
  c = getMonth();
  adr[3] = hexToChar(1,c);
  adr[4] = hexToChar(0,c);
  adr[5] = '-';
  c = getYear();
  adr[6] = hexToChar(1,c);
  adr[7] = hexToChar(0,c); 
  c = getWeekday();
  adr[8] = '#';
  adr[9] = hexToChar(0,c);
  
} 

void readTime(char* adr) {
  char c = getHours();
  adr[0] = hexToChar(1,c);
  adr[1] = hexToChar(0,c);
  adr[2] = ':';
  c = getMinutes();
  adr[3] = hexToChar(1,c);
  adr[4] = hexToChar(0,c);
  adr[5] = ':';
  c = getSeconds();
  adr[6] = hexToChar(1,c);
  adr[7] = hexToChar(0,c);
} 

void readAlarmTime(char* adr) {
  char c = getAlarmHours();
  adr[0] = hexToChar(1,c);
  adr[1] = hexToChar(0,c);
  adr[2] = ':';
  c = getAlarmMinutes();
  adr[3] = hexToChar(1,c);
  adr[4] = hexToChar(0,c);
  adr[5] = ':';
  c = getAlarmSeconds();
  adr[6] = hexToChar(1,c);
  adr[7] = hexToChar(0,c);
  adr[8] = '#';
  c = getAlarmWeekday();
  adr[9] = hexToChar(0,c);
}
 
//++++++++++++++++++++++++++++++++ Control Register Operations +++++++++++++++++++++++++++++++++++++++

// returns 1 if CAF bit is set
char clockAlarm(void) {
  char config = getReg16bit(ACC_CSR,MASK_CAF);
  if(config) return 1; else return 0;
}

// returns 1 if TAF bit is set
char temperatureAlarm(void) {
  char config = getReg16bit(ACC_CSR,MASK_TAF);
  if(config) return 1; else return 0;
}

void stopClockAlarm(void) {
  setAlarmTime("00:00:00#0");
}  

void enableTemperatureAlarm(void) { 
  char config = getReg16bit(ACC_CSR,0xFF00) >> 8;
  config |= CSR_A0;
  setCSReg(config);
}

void enableClockAlarm(void) {
  char config = getReg16bit(ACC_CSR,0xFF00) >> 8;
  config |= CSR_A1;
  setCSReg(config);
}

void disableTemperatureAlarm(void) {
  char config = getReg16bit(ACC_CSR,0xFF00) >> 8;
  config &= ~CSR_A0;
  setCSReg(config);
}

void disableClockAlarm(void) {
  char config = getReg16bit(ACC_CSR,0xFF00) >> 8;
  config &= ~CSR_A1;
  setCSReg(config);
}

char isTemperatureAlarmEnabled(void) {
  char config = getReg16bit(ACC_CSR,0xFF00) >> 8;
  config &= CSR_A0;
  if(config) return 1; else return 0;
}

char isClockAlarmEnabled(void) {
  char config = getReg16bit(ACC_CSR,0xFF00) >> 8;
  config &= CSR_A1;
  if(config) return 1; else return 0;
}

void terminal_handleClockAlarmIRQ(void){
  // CAUTION
  // use clock alarm and temperature in this way or similary
  int h=0;
  if(temperatureAlarm()) {                  // TH threshold alarm
    redOn();                                // optional - switch on red LED
    P1IES |= 0x20;                          // now interrupt on negative edge
  }
  else {
    if(clockAlarm()) {                      // Clock alarm
      yellowOn();                           // optional - switch on yellow LED
      beepOn();                             // optional - turn on beeper
      while(h<10000) h++;                   // optional - delay to hear beep()         
      yellowOff();                          // optional - switch off yellow LED
      beepOff();                            // optional - turn off beeper
    }  
    else {                                  // TL threshold alarm
      redOff();                             // optional - switch off red LED
      P1IES &= 0xDF;                        // now interrupt on positive edge
    }
  }      
  P1IFG &= 0xDF;                            // erase interrupt flag
}


