#include  <msp430x14x.h>
#include "eeprom.h"
#include "sensors.h"
#include "clock.h"

#define SDA_HIGH (P5OUT |= 0x04)
#define SDA_LOW (P5OUT &= 0xFB)
#define SCL_HIGH (P5OUT |= 0x08)
#define SCL_LOW (P5OUT &= 0xF7)

// put start condition on bus
void start(void){
  P5DIR |= 0x0C;                           // ensure: P52(SDA), P53(SCL) output
  SCL_LOW;
  _NOP();
  _NOP();
  SDA_HIGH;
  _NOP();
  _NOP();
  SCL_HIGH;
  _NOP();
  _NOP();
  SDA_LOW;
  _NOP();
  _NOP();
  SCL_LOW;    
  _NOP();
  _NOP();
}

// put stop condition on bus
void stop(void){
  P5DIR |= 0x0C;                           // ensure: P52(SDA), P53(SCL) output
  SCL_LOW;
  _NOP();
  _NOP();
  SDA_LOW;
  _NOP();
  _NOP();
  SCL_HIGH;
  _NOP();
  _NOP();
  SDA_HIGH;
  _NOP();
  _NOP();
  SCL_LOW; 
  _NOP();
  _NOP();
}

// returns zero if failed
// pre:  SCL_LOW
int writeOnBus(char byte) {
  int i, ack;
  for(i=0;i<8;i++) {
    if(byte & 0x80) SDA_HIGH; else SDA_LOW;
    _NOP();
    _NOP();
    SCL_HIGH;
    _NOP();
    _NOP();
    byte = byte << 1;
    SCL_LOW;
    _NOP();
    _NOP();
  }
  // check ack
  P5DIR &= 0xFB;                              // P52(SDA) input
  SCL_HIGH;
  _NOP();
  _NOP();
  if(P5IN & 0x04) ack = 0; else ack = 1;      // test if ack=0, else error
  SCL_LOW;
  _NOP();
  _NOP();
  P5DIR |= 0x04;                              // P52(SDA) output
  _NOP();
  _NOP();
  return ack;
}

// pre:  SCL_LOW
char readFromBus(unsigned char ack) {
  int i;
  char byte=0;
  
  P5DIR &= 0xFB;                              // P52(SDA) input
  
  for(i=0;i<8;i++) {
    byte = byte << 1;                
    SCL_HIGH;
    _NOP();
    _NOP();
    if(P5IN & 0x04) byte |= 0x01; else byte &= 0xFE;
    _NOP();
    SCL_LOW;
    _NOP();
    _NOP();
  }
  P5DIR |= 0x04;                              // P52(SDA) output
  if(ack) SDA_LOW; else SDA_HIGH;
  _NOP();
  SCL_HIGH;
  _NOP();
  SCL_LOW;
  _NOP();
  return byte;
}

// returns zero if one ack was negative (possible after a previous write cycle)
int writeEEPROMByte(char high_adr, char low_adr, char data){

  P5OUT &= 0xEF;                              // disable write protection
  
  do { start(); }                           
  while(!writeOnBus(0xA0));                   // 1010 control, 000 addresse, 0=write             

  if(!writeOnBus(high_adr)) return 0;
  if(!writeOnBus(low_adr))  return 0;
  if(!writeOnBus(data))     return 0;
  _NOP();
  stop();
  P5OUT |= 0x10;                              // enable write protection


  return 1;
}

// waits until ack is positive (possible after a previous write)
char readEEPROMByte(char high_adr, char low_adr){

  char c; 
  do {                                        // wait if writecycle not finished yet
    start();                           
  } while(! writeOnBus(0xA0));                // 1010 control, 000 address, 0=write  --- but only inits address                
                                             
  writeOnBus(high_adr); 
  writeOnBus(low_adr);  
  start();
  writeOnBus(0xA1);                           // 1010 control, 000 address, 1=read     
  c = readFromBus(0);
  stop();

  return c;
}

//writes size bytes into the eeprom
void writeEEPROM(int e_adr, char* r_adr, int size){
  int i;
  int current_e_adr;
  for(i=0;i<size;i++) {
    current_e_adr = e_adr + i;
    while(! writeEEPROMByte(current_e_adr >> 8, current_e_adr, r_adr[i]) ); // wait for eeprom to finish write cycle
  }
}


//reads size bytes from the eeprom, uses readEEPROMByte (slow for large sizes because it uses readEEPROMByte)
void readEEPROM2(int e_adr, char* r_adr, int size){
  int i;
  int current_e_adr;
  for(i=0; i<size; i++){
    current_e_adr = e_adr + i;
    r_adr[i] = readEEPROMByte(current_e_adr >> 8, current_e_adr);
  }
}

//reads size bytes from the eeprom, uses sequential read (faster than readEEPROMByte)
void readEEPROM(int e_adr, char* r_adr, int size){
  unsigned int i, current_e_adr;
  
  do {                                        // wait if writecycle not finished yet
    start();                           
  } while(! writeOnBus(0xA0));                // 1010 control, 000 address, 0=write  --- but only inits address                

  writeOnBus(e_adr >> 8); 
  writeOnBus(e_adr);  
  start();
  writeOnBus(0xA1);                           // 1010 control, 000 address, 1=read     
    
  for(i=0; i<(size-1); i++){
    current_e_adr = e_adr + i;
    r_adr[i] = readFromBus(1);
  }
  r_adr[size-1] = readFromBus(0);                                           
  stop();
}


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// reads the config data from the eeprom to eepromConfig (RAM)
void loadConfig(void){
  readEEPROM(CONFIG_ADDRESS, (char*) &eepromConfig, sizeof(eepromConfig));
  sensorData.id = eepromConfig.id;
}

// writes the config data from eepromConfig (RAM) to the eeprom
void saveConfig(void){
  writeEEPROM(CONFIG_ADDRESS, (char*) &eepromConfig, sizeof(eepromConfig));
  loadConfig(); // reload.. 
}


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
unsigned int eventMemPointer = 0;  // can be 0..EVENTMEM_SIZE

// finds the last eventMemPointer (called for initialization)
void eepromInit(void){
  unsigned int i;                // loopcounter 0..EVENTMEM_SIZE
  struct sensorDataRecord event; // variable to read out eepromevents
  struct timeRecord time;        // time for comparing

  // init with record 0  
  eventMemPointer = 0;
  readEEPROM(EVENTMEM_ADDRESS, (char*) &event, EVENT_SIZE);
  time = event.timeStamp;
  if(event.id == 0xFFFF) return;  // nothing recorded yet, quit here 

  for(i=1; i<EVENTMEM_SIZE; i++){    
    readEEPROM(EVENTMEM_ADDRESS + (i * EVENT_SIZE), (char*) &event, EVENT_SIZE);
    if(event.id == 0xFFFF) break;
    if(compareTimeStamps(&time, (struct timeRecord*) &(event.timeStamp))){
      eventMemPointer = i;
      time = event.timeStamp;        
    }
  }
  eventMemPointer++;
  if(eventMemPointer >= EVENTMEM_SIZE) eventMemPointer=0;   
}


void eeprom_appendEvent(struct sensorDataRecord* event){  
  // TODO: possibly check the eventhistory so not too many events get written into the eeprom
  // for example dont write a microphone event, if one has occurred in the last minute    
  writeEEPROM(EVENTMEM_ADDRESS + (eventMemPointer * EVENT_SIZE), (char*) event, EVENT_SIZE); 
  
  eventMemPointer++;
  if(eventMemPointer >= EVENTMEM_SIZE) eventMemPointer=0;   
}


