#include <io.h>
#include <signal.h>
#include "funk.h"
#include "common.h"
#include "serial.h"
#include "eeprom.h"
#include "ehandler.h"
#include "clock.h"

// sets transceiver into receive mode
// IMPORTANT: when receiving P3SEL for P34(FunkTXD) must be off 
// and a 0 must be output on P34 to achieve good results!!!!! 
void setFunkRx(void){ 
  P3SEL = 0xE0; 
  P5OUT |= 0xC0;     
}

// sets transceiver into ASK mode (amplitude-shift keyed)
void setFunkTxASK(void){
  P3SEL = 0xF0;
  P5OUT |= 0x80;
  P5OUT &= 0xBF;
  wait(400);
}

// sets tranceiver into OOK mode (on-off keyed)
void setFunkTxOOK(void){
  P3SEL = 0xF0;
  P5OUT |= 0x40;
  P5OUT &= 0x7F;
  wait(400);
}

// turns transceiver completely off
void setFunkOff(void){
  P5OUT &= 0x3F;
}

// sends a single byte over radio, this method will block until byte can be written
void sendFunk(unsigned char data){
  while ((IFG1 & UTXIFG0) == 0);        // USART0 TX buffer ready?
    TXBUF0 = data;
}

/* returns transeiver mode - used by terminalprogram
    0 - off   1 - TxASK     2 - TxOOK     3 - RxD       */
char getFunkMode(void) {
   if((P5OUT & 0x80) && (P5OUT & 0x40)) return 0x03;
   if(P5OUT & 0x80) return 0x02;
   if(P5OUT & 0x40) return 0x01;
   return 0x00;
}
// sets tranceiver mode - used by terminalprogram
void setFunkMode(int mode) {
  if(mode==3) setFunkRx();
  else if(mode==2) setFunkTxASK();
  else if(mode==1) setFunkTxOOK();
  else setFunkOff();
}


// set the transmit power (0 <= n <= 100)
void setFunkPower(unsigned char n) {
  int i;
  
  if(n>100) return;
  // Potipegel auf null fahren
  P2OUT &= 0xDF;                                // P25 = 0 (down selected)
  P2OUT &= 0xBF;                                // P26 = 0 (chipselect on)
  for(i=0;i<102;i++) {
     P2OUT &= 0xEF;                             // P24 = 0 (inc)
     P2OUT |= 0x10;    
  }
  // Potipegel==0, Potipegel auf n setzen
  P2OUT |= 0x20;                                // P25 = 1 (up selected)
  for(i=0;i<n;i++) {
    P2OUT &= 0xEF;                              // P24 = 0 (inc)
    P2OUT |= 0x10;    
  }
  P2OUT |= 0x40;                                // P26 = 1 (chipselect off)
  eepromConfig.transmitPower=n;
} 
unsigned char getFunkPower(void) { return eepromConfig.transmitPower; }

void initFunk(void) {
  setFunkRx();  
  setFunkPower(eepromConfig.transmitPower);
  IE1 |= URXIE0;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// ISR and state machine for receiving bytes

volatile unsigned char rxBuf[FUNK_RXBUF_SIZE];    // buffer for receiving
volatile unsigned char rxState = RXSTATE_START1;  // state of receive-statemachine
volatile unsigned int rxBufPos = 0;               // current position in the buffer
volatile unsigned int rxDataLength = 0;           // length of the packet, read out after the sixth byte

#define RXERROR_LENGTH_WRONG   (0x01)  // flags for radio debugging, not fully implemented
#define RXERROR_TOO_LONG       (0x02)
#define RXERROR_WRONG_SBLOCK   (0x04)

#define RX_STANDARD_LENGTH     (50)

volatile unsigned char receivedFlags = 0x00;
volatile unsigned int receivedWrongBytes = 0;

// negates a byte
unsigned char neg(unsigned char x){ return (0xFF - x); }

// send a byte + neg(byte) into the radio
void sendFunkNeg(unsigned char x){
    sendFunk(x);
    sendFunk(neg(x));
}


void funkPacketHandler(void){    
  if(rxState==RXSTATE_FULL){ // packet received    
    beep(50);
    if(eepromConfig.announceFlags & AF_DEBUG_RADIO){
      sendRS232String("GOOD packet received: rxp=");
      sendRS232Int(rxpValue);
      sendRS232String("\tlength=");
      sendRS232Int(rxDataLength);
      sendRS232('\n');
    }
  
    // test for receiver address possibly here!!!

    switch(rxBuf[2]){
      case SENSOR_PACKET:  handleSensorPacket();  break;
      case TIME_PACKET:    setTimeRecord((struct timeRecord*) &rxBuf[6]); break;    
    }
    rxClear();  
  }
  else if(rxState==RXSTATE_ERROR){ // packet with errors 
    if(eepromConfig.announceFlags & AF_DEBUG_RADIO){
      if(receivedFlags & RXERROR_WRONG_SBLOCK) sendRS232String("\tfalse start detected");
      sendRS232String("BAD packet received: rxp=");
      sendRS232Int(rxpValue);
      sendRS232String("\tlength=");
      sendRS232Int(rxDataLength);
      sendRS232String("\twrong bytes=");
      sendRS232Int(receivedWrongBytes);
      sendRS232String("\trxBufPos=");
      sendRS232Int(rxBufPos);
      if(receivedFlags & RXERROR_LENGTH_WRONG) sendRS232String("\twrong length");
      if(receivedFlags & RXERROR_TOO_LONG)  sendRS232String("\ttoo long");
      sendRS232('\n');
    }
    rxClear();    
  }
}


// clears state and pos
void rxClear(void){
  rxState = RXSTATE_WAITING1; 
  rxBufPos = 0;
  rxpValue = 0;
  receivedWrongBytes = 0;
  receivedFlags = 0;
  rxDataLength = 0;
}

volatile unsigned char cdCounter = 0;

unsigned char funkCD(void){
  if((cdCounter > 0) ) return 1;
  else                 return 0; 
}

// receive interrupt for radio
interrupt (UART0RX_VECTOR) funk_rx (void){
  unsigned char rec = RXBUF0;
  
  if(cdCounter==1){ // cdCounter runs out this byte... notice for debugging
    if(rxState==RXSTATE_WAITING1 || rxState==RXSTATE_WAITING2){
      rxState = RXSTATE_START1;
    } else if(rxState==RXSTATE_START1 || rxState==RXSTATE_START2){
      receivedFlags |= RXERROR_WRONG_SBLOCK;
      rxState = RXSTATE_ERROR;
    }  
  }

  if(cdCounter>0) cdCounter--; // decrement cdcounter with each byte
  
  switch(rxState){
    case RXSTATE_ERROR:   // wait until error is processed
      break;        
    
    case RXSTATE_START1:  // waiting for the first startbyte 0x01
      if(rec==0x01) rxState=RXSTATE_START2;  
      break;
      
    case RXSTATE_START2: // waiting for the second startbyte 0x7F
      if(rec==0x7F){
        rxState = RXSTATE_BYTE;
        cdCounter = 15;
      }
      else rxState = RXSTATE_START1;
      break;
      
    case RXSTATE_BYTE: // receive a byte
      rxBuf[rxBufPos] = rec;
      rxBufPos++;
      rxState = RXSTATE_BYTENEG;
      break;
      
    case RXSTATE_BYTENEG: // check the negated byte 
      if(neg(rec) == rxBuf[rxBufPos-1]) rxState = RXSTATE_BYTE;
      
      else {              // negation is wrong!!
        rxState = RXSTATE_BYTE;
        receivedWrongBytes++;
        
        if(rxBufPos==5 || rxBufPos==6){
          receivedFlags |= RXERROR_LENGTH_WRONG;
          rxState = RXSTATE_ERROR;
        }          
      }
    
      // if pos==6 whole header is received => read out the length
      if(rxBufPos==6){
        if(!(receivedFlags & RXERROR_LENGTH_WRONG)){
          rxDataLength = (rxBuf[4] << 8) + rxBuf[5];
          if(rxDataLength > FUNK_RXBUF_SIZE-6){ 
            receivedFlags |= RXERROR_TOO_LONG;
            rxState = RXSTATE_ERROR;
          }
        }
        else {                                        
          rxDataLength = RX_STANDARD_LENGTH;
        }
        cdCounter = rxDataLength * 2 + random(20);          
      }

      // if pos>=6 always check if all bytes have been received
      if(rxBufPos >= 6 && (rxBufPos-6 == rxDataLength) ){
        if(!receivedWrongBytes) rxState = RXSTATE_FULL;
        else                    rxState = RXSTATE_ERROR;
      } 
      break;
      
    case RXSTATE_FULL:  // nothing to do here, wait for someone to get the packet
      break;
      
    case RXSTATE_WAITING1:
       if(rec==0x01) rxState=RXSTATE_WAITING2;  
      break;

    case RXSTATE_WAITING2:
      if(rec==0x7F){
        rxState = RXSTATE_BYTE;
        cdCounter = 15;
      }
      else rxState = RXSTATE_WAITING1;
      break;

    default:
      sendRS232('?'); // shouldnt occur
      break;
      
  }
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// methode for sending packets
unsigned char isSending = 0;
void sendPacket(unsigned char to, unsigned char from, unsigned char type,
                unsigned char num, unsigned int length, char* data){
  unsigned int i;
  
  isSending = 1;
  wait(1000);
  yellowOn();
  
  setFunkTxOOK();

  sendFunk(0xAA); // preamble byte  
  sendFunk(0xAA); // preamble byte
  sendFunk(0xFF); // sync byte
  sendFunk(0x01); // first start byte
  sendFunk(0x7F); // second start byte
  
  sendFunkNeg(to); // to
  sendFunkNeg(from); // from
  sendFunkNeg(type); // type
  sendFunkNeg(num); // num

  sendFunkNeg(length >> 8); // upper 8 bits of length
  sendFunkNeg(length);      // lower 8 bits of length

  for(i=0; i<length; i++) sendFunkNeg(data[i]);  
  
  sendFunk(0xAA); // send two AA bytes after packet for safety
  sendFunk(0xAA);
  yellowOff();
  setFunkRx();
  
  wait(1000);
  isSending = 0;  
}

void broadcastTimeToRadio(struct timeRecord* time){
  sendPacket(BROADCAST, eepromConfig.id, TIME_PACKET, 0x55, sizeof(*time), (char*) time);
}
