//******************************************************************************
// 1780 us bitlength (manchester encoded, so half bitlength of 890 us is important)
//
//   | S  | S  | T  | A4 | A3 | A2 | A1 | A0 | C5 | C4 | C3 | C2 | C1 | C0 |
//   || ?  | ?  | error  | newData || T  | A4 | A3 | A2 || A1 | A0 | C5 | C4 || C3 | C2 | C1 | C0 ||
//
//******************************************************************************

#include <io.h>
#include <signal.h>

#include "recir.h"
#include "sensors.h"
#include "common.h"
#include "serial.h"

#define GREEN_ON      P2OUT &= 0xFD
#define GREEN_OFF     P2OUT |= 0x02

#define BIT75 3282      // 3 quarters after start, 3282 cyc @ 2,4576Mhz = 1335us
#define BIT50 2188      // half of bit length , 2188 cyc @ 2,4576Mhz = 890 us



// only for testing
void testIR(void){
    if(recir_dataAvailable()){
    
      sendRS232String("A: "); sendRS232Int(recir_getAddress());
      sendRS232String(" C: "); sendRS232Int(recir_getCode());
      sendRS232String(" T: "); sendRS232Int(recir_getToggle());
      sendRS232String(" E: "); sendRS232Int(recir_getError());
      sendRS232String("\r\n");    
    
      if(recir_getCode()==RC5_C_VU) redOn();
      if(recir_getCode()==RC5_C_VD) redOff();
      else if(recir_getCode()==RC5_C_REW) yellowToggle();
      
      beep(20);
    }
}

/////////////////////////// forward declarations for "private" functions
void enableFEIRQ(void);
void setErrorBit(void);
void clearErrorBit(void);
void setDataAvailableBit(void);
void clearDataAvailableBit(void);


/////////////////////////////////// Variables
unsigned int ir_pos;             // current position in frame
unsigned int ir_data;            // here a received packet is saved
unsigned int ir_databuffer;      // temporary buffer for receiving
unsigned char ir_temp;           // saves the first half of the manchester bit

////////////////////////// "public" Interface

void recir_init(void){
  P1IES |= 0x04;  // Important for IR-RC5 receive to detect the first FALLING edge
  enableFEIRQ();
}

// gets the code (6 bits)
unsigned char recir_getCode(void){ return (ir_data & 0x003F); }
unsigned char recir_getAddress(void){ return ((ir_data & 0x07C0) >> 6); }
unsigned char recir_getToggle(void){  return ((ir_data & 0x0800) >> 11); }
unsigned char recir_getError(void){ return ((ir_data & 0x2000) >> 13); }

// true if new data available, bit is cleared with this call 
unsigned char recir_dataAvailable(void){  
  if(ir_data & 0x1000){ clearDataAvailableBit(); return 1; }
  else return 0;
}

///////////////////////// "private" part

// initializes and starts the Timer, initializes CCR0 with 75% of RC5 bitlength
// so when called by the first falling edge interrupt ( mid of first start bit )
// the first timer interrupt will occur in the mid of the first half
// of the second start bit 
void startRecIRTimer(void) {
  TBCTL = TBSSEL1 + TBCLR;   // SMCLK, clear TBR
  TBCCTL0 = CCIE;            // CCR0 interrupt enabled, interrupt occurs when timer equals CCR0
  TBCCR0 = BIT75;            // set first CCR0 to 75% of RC5 bitlength
  TBCTL |= MC1;              // Start Timer_A in continuous mode 
}


// enable interrupt for falling edge on recir (begin of rc5-packet), clears IFG before
void enableFEIRQ(void) {
  P1IFG &= 0xFB;  // IMPORTANT to clear interrupt flag as falling edges (may) have occured
                  //  while the interrupt was disabled
  P1IE  |= 0x04;  // enable the interrupt
}

// disables interrupt for falling edge on recir 
void disableFEIRQ(void) {
  P1IE &= 0xFB;   // disable interrupt for P12 ( ReceiveIR )
}

void setErrorBit(void){ ir_data |= 0x2000; }
void clearErrorBit(void) { ir_data &= 0xDFFF; }
void setDataAvailableBit(void){ ir_data |= 0x1000; }
void clearDataAvailableBit(void){ ir_data &= 0xEFFF; }


// Timer B0 interrupt service routine
interrupt(TIMERB0_VECTOR) Timer_B0 (void) {
    if(ir_pos <= 25) {
      if(ir_pos % 2) {                             // odd position
        if(ir_temp && !(P1IN & 0x04)) {           // 1 - 0 -->  write 1 
          ir_databuffer +=1;
          ir_databuffer = ir_databuffer << 1;  
        } 
        else if(!ir_temp && (P1IN & 0x04)) {      // 0 - 1 -->  write 0 
          ir_databuffer = ir_databuffer << 1;  
        } 
        else { 
          setErrorBit();
          if(P1IN & 0x04) ir_databuffer += 1;
          ir_databuffer = ir_databuffer << 1;
        }
      }
      else {                                    // even position
        ir_temp = P1IN & 0x04;
      }
    }
    
    if(ir_pos==25) {                          // end reached
      ir_databuffer = ir_databuffer >> 1; 

      if(!recir_getError() && ( (ir_databuffer & 0x0FFF) != (ir_data & 0x0FFF) ) ){
        ir_data = ir_databuffer;
        setDataAvailableBit();
      } else {
        _NOP();
      }      
    }
    
    if(ir_pos==60) {
      TBCTL &= ~MC1;      // stop IR timer
      TBCCTL0 &= ~CCIE;  
      //GREEN_OFF;
      enableFEIRQ();      // enable IR interrupt
    }
    
  ir_pos++;
  TBCCR0 += BIT50;        // set new interrupt
 
  TBCCTL0 &= 0xFFFE;
}


// IR Interrupt for the falling edge (start of RC5 packet), IRReceiver is on P12
// real interrupt routine, which calls this, is in sensors.c

void handleIRInterrupt(void){
  startRecIRTimer(); //start the timer interrupt
  ir_pos = 0;
  ir_databuffer = 0;
  clearErrorBit();
  
  disableFEIRQ();
  //GREEN_ON;
} 

