#include <io.h>
#include <signal.h>
#include <stdlib.h>

#include "sensors.h"
#include "clock.h"
#include "recir.h"
#include "terminal.h"
#include "common.h"
#include "serial.h"
#include "eeprom.h"
#include "ehandler.h"
#include "funk.h"

#define INTERVAL_5MS      12288       // 12288 cycles @ 2,4756MHz = 5 ms
#define INTERVAL_SENSORS  200         // 200*5ms = 1s

volatile unsigned int micAverage = 2450; // routine initMicrophone should be used to calc this value

// temporary counters where p1irs and the microphone count their events
volatile unsigned int counter[4];
volatile unsigned char buttonPressed=0;
 

// intern variable for calling readOutCounters() every 1s (200 fiveMsTimer-Ticks)
volatile unsigned int nextCounterReadOut = 0;

// for the lightsensor,
// light1, light2 and lightWidth are used for the pulse width measure
// lightFreq is used by the interrupt routin which counts the pulses
volatile unsigned int light1, light2, lightWidth, lightFreq;  

// sets a flag in sensorFlags
void setFlag(unsigned char flag){ sensorData.sensorFlags |= flag; }

// inits the AD to contionously calc the 4 values (extern and battery voltage,
// receivepower and microphone) and generate an interrupt at the
// end of this sequence, the interrupt routine then copies out all values into the ...Value
#define EXTERN    (INCH_3 + SREF_0) // externe Spannung (P63)
#define BATTERY   (INCH_4 + SREF_0) // Batteriespannung (P64)
#define MICRO     (INCH_0 + SREF_0) // mikrofon (P60)
#define RXPOWER   (INCH_5 + SREF_0) // Funk Empfangsleistung (P65) 

void initAD(void){

  ADC12CTL0 = ADC12ON+REFON+REF2_5V+SHT0_6+SHT1_6+MSC;  // Setup ADC12, ref., sampling time
  ADC12CTL1 = SHP+CONSEQ_3+CSTARTADD_0;         // Use sampling timer, repeat-sequence-of-channels

  ADC12MCTL0 = MICRO;             
  ADC12MCTL1 = EXTERN;            
  ADC12MCTL2 = BATTERY;           
  ADC12MCTL3 = RXPOWER + EOS;      
    
  ADC12IE = 0x0008;                     // Enable ADC12IFG.3, Interrupt when sequence is
                                        // completed

  wait(3000);                           // Delay for reference start-up
  
  ADC12CTL0 |= ENC;                     // Enable conversions
  ADC12CTL0 |= ADC12SC;                 // Start conversion
}

// inits Timer A to run continously, CCR1 is used to generate the INTERVAL_5MS
// CCR0 is used to measure the pulse width
void initSensorTimer(void) {
  // init variables
  sensorData.sensorFlags = 0x00;
  sensorData.fiveMsTimer = 0;
  nextCounterReadOut = INTERVAL_SENSORS;
  
  TACTL = TASSEL1 + TACLR;                // select SMCLK (2.4576MHz), clear TAR

  // init ccr0 for capturing light intervals            
  CCTL0 = CM_1 + CCIS_0 + CAP ;           // capture mode: pos edge, capture select: CCIA (P11), capture mode

  // init ccr1 to create the 5 ms interval
  CCTL1 = CCIE;                           // CCR1 interrupt enabled, interrupt occurs when timer equals CCR1
  CCR1 = INTERVAL_5MS;                    // interrupt after 5ms

  TACTL |= MC1;                         // Start Timer_A in continuous mode 

  // Enable port interrupt
  P1IE |= (SENSOR_LIGHT+SENSOR_PIR+SENSOR_VIB); // lightsensor, pirsensor, vibsensor 
}

// captures several microphone adconversions and calculates micAverage
void initMicrophone(void){
  unsigned long int avg = 0;
  unsigned int i = 0;
  
  while(i<10000){ // get 10000 microphone values
    while(!newADValues) nop(); // wait for the ad    
    avg += sensorData.micValue;                        // sum microphone values
    newADValues = 0;
    i++;  
  }
  micAverage = (avg / i);                              // calculate average
}

void initButton(void){ // just enables the interrupt for the button (P27)
  P2IE |= 0x80;
}

void sensor_init(void){ // calls all four initialization functions
  initAD();
  initSensorTimer();
  initMicrophone();
  initButton();
}

void startNewLightWidthMeasure(void){ // starts a new lightWidth measure
  // clear overflow flag and start a new sample by enabling interrupt
  light1 = 0;
  CCTL0 &= ~(COV);
  CCTL0 |= CCIE;
}

//+++++++++++++++++++++++++++ Interrupt routines +++++++++++++++++++++++++++++++

// interrupt help routine, called every 200 Timer_A interrupts ( 1s )
// updates the sensorData values
void readOutCounters(){
  // reads out timestamp to sensorData
  createTimeRecord((struct timeRecord*) &(sensorData.timeStamp));
  // reads out temperature to sensorData
  sensorData.temperature = getReg16bit(ACC_RT,0xFFFF);  
  
  // reads out the lightsensor
  if(P1SEL & 0x02){ // pulswidth was measured, calc frequency and set
    float f = 1.0f/( ((float)lightWidth)*(1.0f/2475600.0f)  );
    sensorData.lightCount =  ((unsigned long int)f);
    startNewLightWidthMeasure();    
  }
  else {  // pulscount was measured
    sensorData.lightCount = lightFreq;
  }
  lightFreq = 0;
  
  if((sensorData.lightCount < eepromConfig.lowerLightLimit) 
    ||(sensorData.lightCount > eepromConfig.upperLightLimit)) setFlag(SENSOR_LIGHT);  

  
  //  IR RC5
 
  if(recir_dataAvailable()){
    sensorData.rc5code = ir_data;
    setFlag(SENSOR_RC5);    
  }
  // button
  if(buttonPressed){
    setFlag(SENSOR_BUTTON);
    buttonPressed=0;
  }

  // pir
  sensorData.pirCount = counter[1];
  counter[1] = 0;
  if(sensorData.pirCount) setFlag(SENSOR_PIR);
  
  // vib  
  sensorData.vibCount = counter[2];
  counter[2] = 0;
  if(sensorData.vibCount) setFlag(SENSOR_VIB);

  // mic
  sensorData.micCount = counter[3];
  counter[3] = 0;
  if(sensorData.micCount) setFlag(SENSOR_MIC);

  // voltages
  if(sensorData.batteryValue < eepromConfig.batteryLimit) setFlag(SENSOR_BATTERYALARM);
  if(sensorData.externValue < eepromConfig.externLimit) setFlag(SENSOR_EXTERNALARM);
    
  nextCounterReadOut += INTERVAL_SENSORS;
}

// light pulswidth measure
interrupt(TIMERA0_VECTOR) Timer_A (void){
  if(light1==0){                       // first capture
   light1 = CCR0; 
  }
  else {                               // second capture
    CCTL0 &= ~(CCIE);                  // turn interrupt off    
    light2 = CCR0;
  
    if(CCTL0 & COV){        // captures overflowed => frequency to high => measure without IR
      CCTL0 &= ~(CCIFG);
      while(!(CCTL0 & CCIFG)); 
      light1 = CCR0;
      
      CCTL0 &= ~(CCIFG);
      while(!(CCTL0 & CCIFG)); 
      light2 = CCR0;
    }
    //calc the width
    if(light2 > light1) lightWidth = light2 - light1; 
    else                lightWidth = (0xFFFF - light1) + light2;
    
    // switch to pulscount if width gets too big
    if(lightWidth > 5000){
      lightFreq = 0;
      P1SEL &= ~(SENSOR_LIGHT);
    }    
  }
}


// every five ms the fiveMsTimer is increased
// every 200 five ms ticks (1s) read out the counters and reset them
interrupt(TIMERA1_VECTOR) Timer_A1 (void){
  switch( TAIV ){
    case  2:                              // CCR1 IRQ: Add Offset to CCR1
      sensorData.fiveMsTimer++;
      if(sensorData.fiveMsTimer == nextCounterReadOut) readOutCounters();
      CCR1 += INTERVAL_5MS;
    break;
  }  
}

// IR for P1, counts the events
interrupt(PORT1_VECTOR) p1_irq (void){
  if(P1IFG & SENSOR_LIGHT){ // light pulscount measure
    lightFreq++;            
    P1IFG &= ~(SENSOR_LIGHT);

    if(lightFreq > 800){    // if pulscount gets too big switch to pulswidth measure
      P1SEL |= SENSOR_LIGHT;
      startNewLightWidthMeasure();
    }    
  }

  if(P1IFG & SENSOR_PIR){
    counter[1]++;
    P1IFG &= ~(SENSOR_PIR);
  } 
  if(P1IFG & SENSOR_VIB){
    counter[2]++;
    P1IFG &= ~(SENSOR_VIB);
  }
  if(P1IFG & 0x04){           // IR receive (recir.c)
    handleIRInterrupt();      // jump into recir.c   
    P1IFG &= ~0x04;           // 
  }
  if(P1IFG & 0x20){           // Clock alarm
    terminal_handleClockAlarmIRQ();
    P1IFG &= ~0x20;
  } 
}


// interrupt for the button
interrupt(PORT2_VECTOR) p2_irq (void){
  buttonPressed = 1;
  P2IFG = 0x00; // button(P27) is the only interrupt generator
}

volatile unsigned int rxpTemp, rxpValue;

volatile unsigned char newADValues = 0;

interrupt(ADC_VECTOR) ADC12ISR (void){

  sensorData.externValue = ADC12MEM1;
  sensorData.batteryValue = ADC12MEM2;
  
  rxpTemp = ADC12MEM3;
  
  if(rxpTemp > eepromConfig.rxReceiveLimit){
    if( (!funkCD()) || (rxState==RXSTATE_WAITING1) || (rxState==RXSTATE_WAITING2)){
      cdCounter = 50;
      rxpValue = rxpTemp;
    } 
  } 
  
  ////// Micro counter, inc counter if abs(micAverage-micValue) > micInterval
  if(!isSending){
    sensorData.micValue = ADC12MEM0; 
    if(abs(micAverage - sensorData.micValue) > eepromConfig.micInterval) counter[3]++;     
  }
  ///////////////      
  newADValues = 1;  
}


