#include <io.h>
#include <signal.h>

#include "serial.h"
#include "common.h"
#include "terminal.h"
#include "funk.h"

#define DELTA 600                       // target DCO = DELTA*(4096) = 2457600 


void sendRS232(unsigned char data){     // sends one byte
  while ((IFG2 & UTXIFG1) == 0);        // USART1 TX buffer ready?
    TXBUF1 = data;
}

// wrappers for sendRS232 (all blocking until all bytes sent)
void sendRS232Chars(char* data, int n){ int i; for(i=0; i<n; i++) sendRS232(data[i]); }
void sendRS232Int(int data){ sendRS232Chars(&intbuf[0], setUnsignedInt(data)); }
void sendRS232LongInt(unsigned long int data){ sendRS232Chars(&longintbuf[0], setUnsignedLongInt(data)); }
void sendRS232String(char* string){
  int i = 0; 
  while(string[i]){
    sendRS232(string[i]);
    i++;
  }
}



// CTS=P16, RTS=P17
unsigned char testCTS(void){ return (P1IN & 0x40); }
void setRTS(void){ P1OUT |= 0x80;}   
void clearRTS(void){ P1OUT &= ~(0x80); }  

void Set_DCO (void){                     // Subroutine to set DCO to selected frequency

  unsigned int Compare, Oldcapture = 0;

  CCTL2 = CCIS0 + CM0 + CAP;            // Define CCR2, CAP, ACLK
  TACTL = TASSEL1 + TACLR + MC1;        // SMCLK, continous mode
  while (1){
    while ((CCTL2 & CCIFG) != CCIFG);   // Wait until capture occured!
    CCTL2 &= ~CCIFG;                    // Capture occured, clear flag
    Compare = CCR2;                     // Get current captured SMCLK
    Compare = Compare - Oldcapture;     // SMCLK difference
    Oldcapture = CCR2;                  // Save current captured SMCLK

    if (DELTA == Compare) break;        // if equal, leave "while(1)"

    else if (DELTA < Compare){           // DCO is too fast, slow it down 
      DCOCTL--;
      if (DCOCTL == 0xFF) BCSCTL1--;     // Did DCO role under?
    }                                    // -> Select next lower RSEL 
    else {                      
      DCOCTL++;
      if (DCOCTL == 0x00) BCSCTL1++;     // Did DCO role over?
                                         // -> Select next higher RSEL      
    }
  }  
  CCTL2 = 0;                              // Stop CCR2 function
  TACTL = 0;                              // Stop Timer_A
}



#define RS232BPS_19200                  // 19200, 38400, 576000 and 115200 supported
void initUart(void)
{  
  unsigned int i;
  BCSCTL1 = 0xa4; // ACLK is devided by 4. RSEL=6 no division for MCLK and SSMCLK. XT2 is off.
  BCSCTL2 = 0x00; // Init FLL to desired frequency using the 32762Hz crystal DCO frquenzy = 2,4576 MHz   
  WDTCTL = WDTPW + WDTHOLD;             // Stop WDT
  BCSCTL1 |= DIVA1 + DIVA0;             // ACLK = LFXT1CLK/8
  for (i = 0xFFFF; i > 0; i--);         // Delay for XTAL to settle
    
  Set_DCO();                            // Software Set DCO for 2,4576MHz  

  //////////////////////////////////////// Funk auf 19200 Bit/s
  UCTL0 = CHAR;                         // 8-bit character
  UTCTL0 = SSEL1;                       // UCLK = SMCLK
  UBR00 = 0x80;                         // 2,457MHz/19200 = 128 -> 0x80
  UBR10 = 0x00;;                        //
  UMCTL0 = 0x00;                        // no modulation
  ME1 |= UTXE0 + URXE0;                 // Enable USART0 TXD/RXD
  //IE1 |= URXIE0;                        // Enable USART0 RX interrupt

  //////////////////////////////////////// RS232
  UCTL1 = CHAR;                         // 8-bit character
  UTCTL1 = SSEL1;                       // UCLK = MCLK

#if defined(RS232BPS_19200)             // RS232 auf 19200 Bit/s
  UBR01 = 0x80;                         // 2,457MHz/19200 = 128 -> 0x80
  UBR11 = 0x00;;                        //
  UMCTL1 = 0x00;                        // no modulation  

#elif defined(RS232BPS_38400)           // RS232 auf 38400 Bit/s
  UBR01 = 0x40;                         // 2,457MHz/38400 = 64 -> 0x40
  UBR11 = 0x00;;                        //
  UMCTL1 = 0x00;                        // no modulation

#elif defined(RS232BPS_57600)           // RS232 auf 57600 Bit/s 
  UBR01 = 0x2A;                         // 2,457MHz/57600 = 42.7 -> 0x2A
  UBR11 = 0x00;;                        //
  UMCTL1 = 0x5B;                        // 
              
#elif defined(RS232BPS_115200)          // RS232 auf 115200 Bit/s
  UBR01 = 0x15;                         // 2,457MHz/115200 = 21.4 -> 0x15
  UBR11 = 0x00;;                        //
  UMCTL1 = 0x4A;                        // 

#else
#warning "define RS232BPS_19200|RS232BPS_38400|RS232BPS_57600|RS232BPS_115200 in serial.c"
#endif

  ME2 |= UTXE1 + URXE1;                 // Enable USART0 TXD/RXD
  IE2 |= URXIE1;                        // Enable USART0 RX interrupt  
  BCSCTL1 &= ~(DIVA1 + DIVA0);          // remove /8 divisor from ACLK again    
}



//-----ISR-Routine----- for receiving bytes
interrupt(UART1RX_VECTOR) usart1_rx (void){
  terminal_handleSerialIRQ();
}

